(function() {
  var COMMANDS = {
    bold: 'bold',
    italic: 'italic',
    underline: 'underline',
    numbers: 'insertOrderedList',
    bullets: 'insertUnorderedList'
  };
  var COMMAND_KEYS_RESET_AFTER_ENTER_UP = ['bold', 'italic', 'underline'];

  var $note = document.getElementById('note');

  var NOTE = {
    reset: function(content) {
      reset(content)
    },
    setPlaceholder: function(placeholder) {
      $note.setAttribute('placeholder', placeholder);
    },
    blurFocus: function() {
      $note.blur();
    },
    toggle: function(cmdKey) {
      document.execCommand(COMMANDS[cmdKey], true, null);
      updateState();
    }
  };
  window.NOTE = NOTE;

  document.addEventListener('selectionchange', updateState);
  var clickedY = 0;
  $note.addEventListener('click', function(e) {
    updateState();
    clickedY = e.y;
  });

  window.onresize = function() {
    var windowHeight = document.firstElementChild.offsetHeight;
    if (clickedY > windowHeight) {
      scrollTo(0, clickedY);
    }
  };

  $note.addEventListener('input', updateContent);

  $note.addEventListener('keyup', function(e) {
    var KEY_LEFT = 37;
    var KEY_RIGHT = 39;
    var KEY_ENTER = 13;
    var KEY_DELETE = 8;

    switch (e.which) {
      case KEY_LEFT:
      case KEY_RIGHT:
        updateState();
        break;
      case KEY_DELETE:
        _handleDeleteUp();
        break;
      case KEY_ENTER:
        _handleEnterUp();
        break;
      default:
        break;
    }

    function _handleDeleteUp() {
      updateState();
      updateContent();
    }

    function _handleEnterUp() {
      COMMAND_KEYS_RESET_AFTER_ENTER_UP
        .forEach(_resetCommandByKey);
      updateState();
    }

    function _resetCommandByKey(key) {
      if (isCommandActive(key)) {
        document.execCommand(COMMANDS[key]);
      }
    }
  });

  function updateContent() {
    window.location.href = 'note-content://' + encodeURIComponent(getContent());
  }

  function updateState() {
    var keys = Object.keys(COMMANDS);
    var activeCommands = keys
        .filter(isCommandActive)
        .join(',');
    setTimeout(function() {
      window.location.href = 'note-state://' + encodeURIComponent(activeCommands);
    }, 0);
  }

  function isCommandActive(key) {
    return document.queryCommandState(COMMANDS[key]);
  }

  function reset(content) {
    if (content.length > 0 && content[0] === '{') {
      var note = JSON.parse(content);
      var html = '';

      if (note.ops) {
        html = opsToHtml(note.ops.ops);
      } else if (note.html) {
        html = xmind8ToHtml(note.html.content.paragraphs);
      } else {
        html = note.plain.content;
      }

      $note.innerHTML = html;
    } else {
      $note.innerText = content;
    }
  }

  function getContent() {
    if ($note.innerText.trim() === '') {
      return '';
    }

    return JSON.stringify({
      ops: {
        ops: toOps($note)
      },
      html: {
        content: {
          paragraphs: toXmind8($note)
        }
      },
      plain: {
        content: $note.innerText
      }
    });
  }

  function opsToHtml(ops) {
    var lastLineStart = 0;
    var lastListType = 'none';
    var html = '';

    ops.forEach(function (it, index) {
      if (isListNode(it)) {
        var type = it.attributes.list;

        html = html.substring(0, lastLineStart) +
          getLastListSuffix(type, lastListType) +
          getListPrefix(type, lastListType) +
          wrapLi(html.slice(lastLineStart));

        if (index === ops.length - 1) {
          html += getListSuffixBy(type);
        }

        lastLineStart = html.length;
        lastListType = type;
      } else {
        if (lastListType !== 'none' && it.insert.startsWith('\n')) {
          var suffix = getListSuffixBy(lastListType);

          html = html.substring(0, lastLineStart) +
            suffix +
            html.slice(lastLineStart);

          lastListType = 'none';
          lastLineStart = html.lastIndexOf(suffix) + 1;
        }

        html += wrap(it);

        if (it.insert.includes('\n')) {
          lastLineStart = html.lastIndexOf('\n') + 1;
        }
      }
    });

    return html.replace(/\n/g, '<br>');

    function isListNode(node) {
      return node.attributes && node.attributes.list;
    }

    function getListPrefix(type, lastType) {
      var ret = '';
      if (type !== lastType) {
        ret = type === 'ordered' ? '<ol>' : '<ul>';
      }

      return ret;
    }

    function getLastListSuffix(type, lastType) {
      var ret = '';
      if (lastType !== 'none' && type !== lastType) {
        ret = getListSuffixBy(lastType);
      }
      return ret;
    }

    function getListSuffixBy(type) {
      return type === 'ordered' ? '</ol>' : '</ul>';
    }

    function wrapLi(content) {
      return '<li>' + content + '</li>';
    }
  }

  function xmind8ToHtml(paragraphs) {
    var html = '';

    paragraphs.forEach(function (paragraph, index) {
      paragraph.spans.forEach(function (span) {
        if (span.text && span.text !== '') {
          html += wrap(span);
        }
      });

      if (index !== paragraphs.length - 1) {
        html += '<br>';
      }
    });

    return html;
  }

  function wrap(node) {
    var surrounds = [{
      key: 'fo:font-weight',
      value: 'bold',
      prefix: '<b>',
      suffix: '</b>'
    }, {
      key: 'fo:font-style',
      value: 'italic',
      prefix: '<i>',
      suffix: '</i>'
    }, {
      key: 'fo:text-decoration',
      value: 'underline',
      prefix: '<u>',
      suffix: '</u>'
    }];

    var styles = node.attributes ||
        (node.style && node.style.properties) || {};
    var ret = '';
    surrounds.forEach(function (it) {
      if (hasStyle(styles, it)) {
        ret += it.prefix;
      }
    });
    ret += escape(node.insert || node.text);
    surrounds.forEach(function (it) {
      if (hasStyle(styles, it)) {
        ret += it.suffix;
      }
    });

    return ret;

    function hasStyle(styles, surround) {
      return styles[surround.value] ||
        (styles[surround.key] && styles[surround.key].includes(surround.value));
    }

    function escape(text) {
      return !text ? '' : text
        .replace(/&/g, '&amp;')
        .replace(/</g, '&lt;')
        .replace(/>/g, '&gt;');
    }
  }

  function toOps(el) {
    var ops = [];

    var texts = textNodesUnder(el);

    texts.forEach(function (node) {
      ops.push({
        attributes: getAttrs(node),
        insert: isLiEnd(node) ? '\n' : node.data
      });
    });

    return ops;

    function getAttrs(node) {
      var attrs = {};

      if (isLiEnd(node)) {
        attrs.list = getLiType(node);
      } else if (node.data !== '\n') {
        if (isUnderline(node)) {
          attrs.underline = true;
        }
        if (isBold(node)) {
          attrs.bold = true;
        }
        if (isItalic(node)) {
          attrs.italic = true;
        }
      }

      return attrs;
    }
  }

  function toXmind8(el) {
    var paragraphs = [];
    var paragraph = {};
    var spans = [];
    var nodes = textNodesUnder(el);

    nodes.forEach(function (node, index) {
      var properties = getProperties(node);
      var span = {
        text: isLiEnd(node) || node.data === '\n' ? '' : node.data
      };
      if (Object.keys(properties).length > 0) {
        span.style = {
          type: 'text',
          properties: properties
        };
      }

      spans.push(span);
      if (isLiEnd(node) || node.data === '\n' || index === nodes.length - 1) {
        if (spans.length > 1) {
          spans.pop();
        }
        paragraph.spans = spans;
        paragraphs.push(paragraph);
        paragraph = {};
        spans = [];
      }
    });

    return paragraphs;

    function getProperties(node) {
      var properties = {};

      if (!isLiEnd(node) && node.data !== '\n') {
        if (isUnderline(node)) {
          properties['fo:text-decoration'] = 'underline';
        }
        if (isBold(node)) {
          properties['fo:font-weight'] = 'bold';
        }
        if (isItalic(node)) {
          properties['fo:font-style'] = 'italic';
        }
      }

      return properties;
    }
  }

  function textNodesUnder(node) {
    var all = [];
    for (node = node.firstChild; node; node = node.nextSibling) {
      if (node.nodeType === 3) {
        all.push(node);
      } else if (node.nodeName === 'BR') {
        all.push(document.createTextNode('\n'));
      } else {
        all = all.concat(textNodesUnder(node));
        if (node.nodeName === 'LI') {
          var type = node.parentElement.nodeName === 'UL' ? 'bullet' : 'ordered';
          all.push(createLiEndNode(type));
        }
      }
    }
    return all;
  }

  var LI_END = 'xmind-li-end-flag:';
  function createLiEndNode(type) {
    return document.createTextNode(LI_END + type);
  }

  function getLiType(node) {
    return node.data.slice(LI_END.length);
  }

  function isLiEnd(node) {
    return node.data.startsWith(LI_END);
  }

  function isUnderline(node) {
    var parent = node.parentElement;
    if (!parent) {
      return false;
    } else {
      var style = getComputedStyle(parent);
      if (style.getPropertyValue('text-decoration-line') === 'underline') {
        return true;
      } else {
        return isUnderline(parent);
      }
    }
  }

  function isBold(node) {
    var style = getComputedStyle(node.parentElement);
    var weight = style.getPropertyValue('font-weight');

    var ret = false;

    if (parseInt(weight) > 500) {
      ret = true;
    } else if (weight.toLowerCase().includes('bold')) {
      ret = true;
    }

    return ret;
  }

  function isItalic(node) {
    var style = getComputedStyle(node.parentElement);
    var fontStyle = style.getPropertyValue('font-style');
    return fontStyle.toLowerCase() === 'italic';
  }

  function polyfill() {
    if (!String.prototype.includes) {
      String.prototype.includes = function(search, start) {
        if (typeof start !== 'number') {
          start = 0;
        }

        if (start + search.length > this.length) {
          return false;
        } else {
          return this.indexOf(search, start) !== -1;
        }
      };
    }

    if (!String.prototype.startsWith) {
      Object.defineProperty(String.prototype, 'startsWith', {
        value: function(search, pos) {
            pos = !pos || pos < 0 ? 0 : +pos;
            return this.substring(pos, pos + search.length) === search;
        }
      });
    }
  }

  polyfill();
})();
